/*
 * Driver for the SPCP825 based Optocoupler / Sunplus RS232C to USB Adapter
 *
 * Copyright (C) 2007  Martin Schewe  <multimeter@schewe.com>
 * Modified      2014-07-01	Anonymous	Modified to compile with MinGW with libusb-win32
 
 multimeter-0.1/libusb-win32-bin-1.2.6.0/bin/x86
> testlibusb.exe

Dev #0: 0000 - 0000

  Dev #1: 04FC - USB to RS232C Adapter
 
 
 * You always find the most recent version on http://multimeter.schewe.com.
 * $Id: optocoupler.c 3 2007-11-14 22:09:43Z ms $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
//#include <usb.h>
#include "libusb-win32-bin-1.2.6.0/include/lusb0_usb.h"


#define VERSION		"0.1"

#define USB_VENDOR	0x04fc
#define USB_PRODUCT	0x0201
#define USB_EP		(0x81-0)
#define TIMEOUT		1000
#define BUF_LEN		1024
#define PACKET_LEN	14


enum output_format	{ BIN, HEX, RAW };


/*
 * Show a usage message
 */

void usage()
{
	fprintf(stderr,
		"Optocoupler Driver for the V&A VA18B Multimeter %s\n"
		"Copyright (C) 2007  Martin Schewe  <multimeter@schewe.com>\n\n"

		"Usage: optocoupler [OPTION]...\n\n"

		"-b	binary output (default)\n"
		"-h	hexadecimal output\n"
		"-r	raw output\n"
		"-n	filter serially numbered nibbles\n",
		VERSION);

	exit(EXIT_FAILURE);
}


/*
 * Search for the first supported device on the USB bus and open it
 */

struct usb_dev_handle *oc_open()
{
	struct usb_bus    *bus;
	struct usb_device *dev;

	usb_init();
	
	usb_set_debug( 3 );
	
	usb_find_busses();
	usb_find_devices();

	fprintf(stderr, "scanning USB devices\n");
	
	for (bus = usb_busses; bus; bus = bus->next)
	{
		for (dev = bus->devices; dev; dev = dev->next)
		{
			
			fprintf(stderr, "Device: Vendor: 0x%04X Product: 0x%04X\n", dev->descriptor.idVendor, dev->descriptor.idProduct);
			
			if (dev->descriptor.idVendor  == USB_VENDOR &&
			    dev->descriptor.idProduct == USB_PRODUCT)
			{
				fprintf(stderr, "found device\n");
				
				
				/* Loop through all of the configurations */
				int c, i, a,n;
				for (c = 0; c < dev->descriptor.bNumConfigurations; c++) 
				{
					/* Loop through all of the interfaces */
					for (i = 0; i < dev->config[c].bNumInterfaces; i++) 
					{
						/* Loop through all of the alternate settings */
						for (a = 0; a < dev->config[c].interface[i].num_altsetting; a++) 
						{
							/* Loop through all of the Endpoints settings */
							for (n = 0; n < dev->config[c].interface[i].altsetting[a].bNumEndpoints; n++) 
							{
								fprintf(stderr, "Endpoint address: 0x%02X\n", dev->config[c].interface[i].altsetting[a].endpoint[n].bEndpointAddress );
								fprintf(stderr, "Max packet size: %d\n", dev->config[c].interface[i].altsetting[a].endpoint[n].wMaxPacketSize );
							}
						}
					}
				}
				
				
				
				
				
				
				
				return usb_open(dev);
			}
		}
	}
	return NULL;
}


/*
 * Claim the device and setup the timeing (XXX)
 */

int oc_init(struct usb_dev_handle *devh)
{

	fprintf(stderr, "claim interface, send control messages\n");
	
	int ret = 0;
	
	ret |= usb_claim_interface(devh, 0);
	fprintf(stderr, "Status: %d, Error: %s\n", ret, usb_strerror() );
	
	ret |= usb_control_msg(devh, USB_TYPE_VENDOR | USB_RECIP_DEVICE, 0x21, 0x303, 0, NULL, 0, TIMEOUT);
	fprintf(stderr, "Status: %d, Error: %s\n", ret, usb_strerror() );
	
	ret |= usb_control_msg(devh, USB_TYPE_VENDOR | USB_RECIP_DEVICE, 0x22, 0x100, 0, NULL, 0, TIMEOUT);
	fprintf(stderr, "Status: %d, Error: %s\n", ret, usb_strerror() );
	
	ret |= usb_control_msg(devh, USB_TYPE_VENDOR | USB_RECIP_DEVICE, 0x21, 0x303, 0, NULL, 0, TIMEOUT);
	fprintf(stderr, "Status: %d, Error: %s\n", ret, usb_strerror() );
	
	return ret;
}


/*
 * Print the raw/hex/bin data to stdout
 */

void print_data (enum output_format fmt, unsigned char data, int is_a_nibble)
{
	switch (fmt) {
		case RAW:
			printf("%c", data);
			break;
		case HEX:
			printf(is_a_nibble ? "%01hhx" : "%02hhx", data);
			break;
		case BIN:
			if (!is_a_nibble)
				printf("%i%i%i%i", (data >> 7) & 1,
						   (data >> 6) & 1,
						   (data >> 5) & 1,
						   (data >> 4) & 1);

			printf("%i%i%i%i", (data >> 3) & 1,
					   (data >> 2) & 1,
					   (data >> 1) & 1,
					   (data     ) & 1);
			break;
	}

}

/*
 * Parse arguments and write the (possibly filtered) data to stdout
 */

int main(int argc, char **argv)
{
	enum output_format	output_fmt = BIN;
	int			snn_filter = 0;

	fprintf(stderr, "starting...\n");
	
	int opt;
	while ((opt = getopt(argc, argv, "bhrn")) != -1) {
		switch (opt) {
			case 'b':
				output_fmt = BIN;
				break;
			case 'h':
				output_fmt = HEX;
				break;
			case 'r':
				output_fmt = RAW;
				break;
			case 'n':
				snn_filter = 1;
				break;
			default:
				usage();
		}
	}
	fprintf(stderr, "cli options done\n");
	
	fprintf(stderr, "open device...\n");
	struct usb_dev_handle* devh = oc_open();
	if (!devh) {
		fprintf(stderr, "open failed\n");
		fprintf(stderr, "Error: %s\n", usb_strerror() );
		//usb_exit();
		exit(EXIT_FAILURE);
	}
	
	fprintf(stderr, "setup device...\n");
	if (oc_init(devh)) {
		fprintf(stderr, "init failed\n");
		fprintf(stderr, "Error: %s\n", usb_strerror() );
		usb_close(devh);
		//usb_exit();
		exit(EXIT_FAILURE);
	}
	
	
	
	char buf[BUF_LEN];
	
	int x;
	for( x = 0; x<BUF_LEN; x++)
	{
		buf[x] = 0;
	}
	
	char packet[PACKET_LEN];
	//int transferred;

	fprintf(stderr, "device opened, discarding buffer...\n");
	
	// discard stale buffer
	int ret = usb_bulk_write(	devh, 			// dev_handle
														USB_EP,			// endpoint
														buf,				// data
														BUF_LEN,		// length
														1						// timeout
														);
	if(ret < 0)
		{
			fprintf(stderr, "USB bulk write return value: %d\n", ret);
			//fprintf(stderr, "USB bulk write transferred: %d of %d\n", transferred, BUF_LEN);
			//fprintf(stderr, "Error name: %s\n", usb_error_name(ret) );
			fprintf(stderr, "Error: %s\n", usb_strerror() );
			usb_close(devh);
			//usb_exit();
			exit(EXIT_FAILURE);
		}

	fprintf(stderr, "entering loop..(abort with CTRL+C)\n");
	int next_nibble = 1;
	while (1) {   
		ret = usb_bulk_write(devh, USB_EP, buf, BUF_LEN, TIMEOUT);
		if(ret < 0)
		{
			fprintf(stderr, "USB bulk write return value: %d\n", ret);
			fprintf(stderr, "Error: %s\n", usb_strerror() );
			usb_close(devh);
			//usb_exit();
			exit(EXIT_FAILURE);
			//break;
		}
		int i;
		for (i = 0; i < ret; i++) {
			if (snn_filter) {
				int nibble_number = buf[0] >> 4 & 0xf;
				int nibble_data   = buf[0]      & 0xf;

				// eliminate broken packets
				if (nibble_number != next_nibble) {
					next_nibble = 1;
					continue;
				}

				packet[next_nibble++ - 1] = nibble_data;

				if (nibble_number == PACKET_LEN) {
					next_nibble = 1;
					int j;
					for (j = 0; j < PACKET_LEN; j++)
						print_data(output_fmt,
							   packet[j], 1);
					printf("\n");
				}
			} else
				print_data(output_fmt, buf[i], 0);
		}
		fflush(stdout);
	}
	
	fprintf(stderr, "exiting.. \n");
	usb_close(devh);
	//usb_exit();
	return EXIT_SUCCESS;
}
